#include <kgi/kgi.h>
#include <kgi/module.h>
#include <kgi/kgi_commands.h>

#include "3dfx/banshee.h"


#define	DRIVER_NAME	"3DFX Banshee DAC driver"
#define	DRIVER_REV	"$Revision: 1.1 $"

static void 
dac_set_clut(struct kgi_display *dpy, kgi_clut *clut,
	     kgi_uint first, kgi_uint cnt)
{
  GET_PRIVATE;
  kgi_color *color ; 
  kgiu16 i ; 
  ASSERT(cnt);
  
  color = &clut->data[first] ; 
  for (i=0;i<cnt;i++)
    {
      writel(first+i,IOMEMBASE+DACADDR);
      writel(
	     ( (color->r & 0xFF00) << 8) | 
	     (color->g & 0xFF00) | 
	     (color->b >> 8),
	     IOMEMBASE+DACDATA 
	    );
      color++ ; 
    }

}

static void 
dac_get_clut(struct kgi_display *dpy,
	     kgi_clut *clut, kgi_uint first, kgi_uint cnt)
{
  GET_PRIVATE;
  kgi_color *color ; 
  kgiu16 i ; 
  ASSERT(cnt);

  color = &clut->data[first] ; 
  for (i=0;i<cnt;i++)
    {
      kgiu32 c ; 
      writel(first+i,IOMEMBASE+DACADDR);
      c=readl(IOMEMBASE+DACDATA) ;
      color->r = (c >> 16 ) * 0x101 ; 
      color->g = ((c >> 8) & 0xFF )  * 0x101 ; 
      color->b = (c & 0xFF)  * 0x101 ; 
      color++ ; 
    }
}
static struct kgi_ramdac ramdac = 
{
  "3DFX",			/* Manufacturer		*/
  "Banshee",		/* Model		*/
  GM_TEXT16 | GM_8BIT |	/* Supported modes	*/
  GM_16BIT | GM_24BIT | GM_32BIT,
  FL_DAC_8BIT,		/* Flags		*/
  { 2048, 2048 },		/* Maximum resolution	*/
  NULL,			/* Private data hook	*/
  DAC_8BIT_PB,  		/* Device type		*/
  {0,  BANSHEE_MAX_CLOCK},	/* Normal mode		*/
  {0,         0 },	/* 2xclk mode		*/
  {
    {0,  80000000 },	/* KGIGT_TEXT16	*/
    {0,         0 },	/* KGIGT_TEXT32	*/
    {0,         0 },	/* KGIGT_1BIT	*/
    {0,         0 },	/* KGIGT_4BIT	*/
    {0, BANSHEE_MAX_CLOCK },	/* KGIGT_8BIT	*/	
    {0,         0 },	/* KGIGT_15BIT	*/
    {0, BANSHEE_MAX_CLOCK },	/* KGIGT_16BIT	*/
    {0, BANSHEE_MAX_CLOCK },	/* KGIGT_24BIT	*/
    {0, BANSHEE_MAX_CLOCK }	/* KGIGT_32BIT	*/
  }
};



/* GGI interface */

int kgim_ramdac_init(struct kgi_card *card, struct kgi_display *dpy)
{
  GET_PRIVATE;
  kgi_uint speed;
  NOTICE(DRIVER_NAME " rev. " DRIVER_REV);
  
  /* Set DAC speed dependent on revision */
  switch(TYPE)
    {
    case banshee_3dfx:
      speed = BANSHEE_MAX_CLOCK;
      break ;
    default:
      /* higher for VOODOO3... ? */
      speed = BANSHEE_MAX_CLOCK;
      break ; 
    }

  CHIPSET_INFO.dclk.max =
    ramdac.normal.max =
    ramdac.clock[KGIGT_8BIT].max =
    ramdac.clock[KGIGT_16BIT].max =
    ramdac.clock[KGIGT_24BIT].max =
    ramdac.clock[KGIGT_32BIT].max = speed;

  NOTICE("Maximum allowed dot-clock : %d.000 Mhz",speed/1000000);
  
  return EOK;
}

void kgim_ramdac_done(struct kgi_display *dpy)
{
  NOTICE(" RAMDAC_done.");
}

int kgim_ramdac_check_mode (struct kgi_display *dpy, struct kgi_mode *mode,
			    enum kgi_tm_cmd *cmd)
{
  uint newclock = mode->dclk;
  DEBUG1(" RAMDAC_check_mode %d.",*cmd);

  if (!((1<<mode->request.graphtype)&ramdac.modes))
    {
      NOTICE("  couldn't handle graphtype %i",mode->request.graphtype);
      DEBUG2("couldn't handle graphtype %i", 
	     mode->request.graphtype);
      return -E(RAMDAC,MODE_CHECK);
    }

  switch (*cmd) {

  case CMD_PROPOSE:
    DEBUG1(" => CMD_PROPOSE");

    mode->lclk.mul = mode->clk.mul = 1;
    mode->lclk.div = mode->clk.div = 1;
    return EOK;

  case CMD_LOWER:
    DEBUG1(" => CMD_LOWER");

    if (ramdac.clock[mode->request.graphtype].max < newclock) {
      newclock = ramdac.clock[mode->request.graphtype].max;
    }
    mode->dclk = newclock;
    return EOK;

  case CMD_RAISE:	
    DEBUG1(" => CMD_RAISE");

    if ((newclock > ramdac.clock[mode->request.graphtype].max) ||
	(newclock > ramdac.normal.max)) {

      DEBUG2("DCLK too high (%i Hz).", newclock);
      return -E(RAMDAC,CLK_HIGH_LIMIT);
    }

    mode->dclk = newclock;
    return EOK;

  case CMD_CHECK:
    DEBUG1(" => CMD_CHECK");

    if ((mode->lclk.mul != 1) || (mode->lclk.div != 1) ||
	(mode->clk.mul != 1) || (mode->clk.div != 1)) {

      return -E(RAMDAC,MODE_CHECK);
    }

    if ((mode->dclk > ramdac.clock[mode->request.graphtype].max) ||
	(mode->dclk > ramdac.normal.max)) {
			
      return -E(RAMDAC,CLK_HIGH_LIMIT);
    }

    return EOK;

  default: 

    INTERNAL_ERROR;
    return -E(RAMDAC,BOGUS);
  }
}



void kgim_ramdac_set_mode(struct kgi_display *dpy,struct kgi_mode *mode)
{
  dpy->set_clut = dac_set_clut; 
}

int kgim_ramdac_command(struct kgi_graphic *graph, kgi_uint cmd, kgi_uint arg)
{
  /* This may set 
   * -  Hardware cursor
   * -  CLUT (or gamma correction table)
   */
  kgi_clut clut;

  switch (cmd) { 
	  
  case RAMDAC_GETINFO:
    graph->io_res = &ramdac;
    return EOK;
	  
  case RAMDAC_SETINFO:
    if (suser()) {
      memcpy(&ramdac, graph->io_buf, sizeof(ramdac));
      return EOK;
    } 
    else {
      return -EPERM;
    }
  case RAMDAC_SETCLUT:
    clut.size = 256;
    clut.data = (kgi_color *)(graph->io_buf);
    dac_set_clut(graph->dev.dpy, &clut, 0, 256); 
    return 0;
	  
  case RAMDAC_GETCLUT:
    clut.size = 256;
    clut.data = (kgi_color *)(graph->io_buf);
    graph->io_res = graph->io_buf;
    dac_get_clut(graph->dev.dpy, &clut, 0, 256); 
    return 0;
  }
  return -E(RAMDAC, NOSUP_ALWAYS_CANT);  /* The ramdac can't do this */
}

